;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; SYNCTSR   - TSR program that stays resident and tries to sync time with
;             the host machine running the NTVDM in order to prevent clock 
;             drifts. On some strange machines, this problem seems to occur,
;             no idea, why....
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;     Version 1.0                                                2017/04/05
;
;     Copyright (c) 2017 Dipl.-Ing. (FH) Ludwig Ertl
;     ludwig.ertl@dose.0wnz.at
;     http://www.waldbauer.com/tmp/reference.php
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
.model tiny

BOP_DOS		EQU	50H 		; BOP for DOS functions
SVC_DEMDATE16	EQU	30H 		; Service for system DATE as DOS date

BOP macro BopNumber
    db  0C4h, 0C4h, BopNumber
endm

svc macro   func
    BOP BOP_DOS
    db  func
endm

CODE SEGMENT BYTE PUBLIC 'CODE'
ASSUME CS:CODE, DS:CODE, ES:CODE
ORG 100h

; Here we go -----------------------------[EIP]-------------------------------
start:
	JMP	install			; Install TSR

	TMR_ACT DB 0			; Is it desirable to adjust time in INT28h?
	TICK_91	DB 91			; Countdown until it is desirable to check

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; INT 1C handler ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
NEW_1CINT:
	PUSH DS				; Save DS, DS=CS
	PUSH CS
	POP DS

	DEC	TICK_91			; Decrement until ~5 seconds passed
	JNZ	GO_OLD1C
	INC	TMR_ACT			; Set timer active, for reentrance check
	MOV	TICK_91, 91		; Reset countdown

GO_OLD1C:				; Go to original INT 1Ch handler
	POP DS
	DB 0EAh  ; JMPF
	OLD_1CINT DD 0

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; INT 28 handler ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
NEW_28INT:
	PUSH DS				; Save DS, DS=CS
	PUSH CS
	POP DS
	CMP	TMR_ACT, 0
	JNZ	DO_TICK
GO_OLD28:				; Go to original INT 28h handler
	POP DS
	DB 0EAh  ; JMPF
	OLD_28INT DD 0

DO_TICK:
	MOV	TMR_ACT, 0		; Reset to 0, prevent reentrance
	PUSH AX				; Save registers
	PUSH BX
	PUSH CX
	PUSH DX

	MOV	AH, 2Ch			; Get time from NTVDM's DOS 
	INT	21h
	MOV	AL, CL			; Convert time to neede DOS time fmt -> CX
	SHL	CH, 3			; hour
	SHR	AL, 3
	OR	CH, AL
	SHL	CL, 5
	SHR	DH, 1
	OR	CL, DH
	SVC	SVC_DEMDATE16		; Get current system time as DOS-Date -> DX
	MOV	AX, DX			; HHHHHMMMMMMSSSSS  (<<11 | (&3f)<<5| (&1f)
	SUB	DX, CX			; More then 2 seconds difference?
	CMP	DX, 2
	JG	SYNC_TIME
	CMP 	DX, -2
	JGE	TIME_OK

SYNC_TIME:
	MOV	CX, AX			; Then let's sync!
	MOV	DH, CL			; second
	AND	DH, 1Fh
	SHL	DH, 1
	XOR	DL, DL			; 1/100 seconds
	SHR	CX, 3			; Hours
	SHR	CL, 2			; Minutes
	MOV	AH, 2Dh			; Set time
	INT	21h

TIME_OK:
	POP DX				; Restores registers
	POP BX
	POP CX
	POP AX
	JMP	GO_OLD28		; Back to original handler


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Installation ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
install:
	MOV	AX, 351Ch		; Get old INT 1Ch handler
	INT	21h

	CMP	BX, OFFSET NEW_1CINT	; Our function starts at NEW_1CINT
	JNZ	inst_ok
	CMP	ES:[BX], 0E1EH		; Our function starts with PUSH DS...
	JNZ	inst_ok

	MOV	DX, OFFSET ERR_INSTLD	; Seems we are alread installed
	MOV	AH, 09h
	INT	21h

	MOV	AX, 4C01h		; Already installed -> Bye!
	INT	21h

inst_ok:
	MOV	WORD PTR [OLD_1CINT], BX	; Save original INT1C handler
	MOV	WORD PTR [OLD_1CINT + 2], ES

	MOV	AX, 251Ch		; Install new handler for timer
	MOV	DX, OFFSET NEW_1CINT
	INT	21h

	MOV	AX, 3528h		; Get old INT 28h handler (DOS idle is safe
	INT	21h			; for calling INT 21h like we want to)

	MOV	WORD PTR [OLD_28INT], BX	; Save original INT28 handler
	MOV	WORD PTR [OLD_28INT + 2], ES

	MOV	AX, 2528h		; Install new handler
	MOV	DX, OFFSET NEW_28INT
	INT	21h

	MOV	DX, OFFSET install	; Terminate and stay resident
	INT	27h

; -- Messages
ERR_INSTLD	db 'Seems like SyncTSR is already installed, abort.',13,10,'$'



Code ends
End start
